#include <cfloat>
#include <engine/collision.h>
#include <J3ML/Geometry.h>
#include <J3ML/Geometry/AABB.h>
#include <J3ML/Geometry/Triangle.h>

bool Collision::vertexInsideAxisAlignedBoundingBox(const Vector3& position, const Collider& boundingBox) {
    Vector3 minimum = {boundingBox.vertices.front().x, boundingBox.vertices.front().y, boundingBox.vertices.front().z};
    Vector3 maximum = {boundingBox.vertices.back().x, boundingBox.vertices.back().y, boundingBox.vertices.back().z};

    return (
            position.x >= minimum.x && position.x <= maximum.x &&
            position.y >= minimum.y && position.y <= maximum.y &&
            position.z >= minimum.z && position.z <= maximum.z
    );
}

bool Collision::vertexInsideOrientedBoundingBox(const Vector3& position, const Collider& boundingBox) {

    Vector3 minimum = {FLT_MAX, FLT_MAX, FLT_MAX};
    Vector3 maximum = {-FLT_MAX, -FLT_MAX, -FLT_MAX};

    for (const auto& vertex : boundingBox.vertices) {
        if (vertex.x < minimum.x) minimum.x = vertex.x;
        if (vertex.y < minimum.y) minimum.y = vertex.y;
        if (vertex.z < minimum.z) minimum.z = vertex.z;

        if (vertex.x > maximum.x) maximum.x = vertex.x;
        if (vertex.y > maximum.y) maximum.y = vertex.y;
        if (vertex.z > maximum.z) maximum.z = vertex.z;
    }

    return (
            position.x >= minimum.x && position.x <= maximum.x &&
            position.y >= minimum.y && position.y <= maximum.y &&
            position.z >= minimum.z && position.z <= maximum.z
    );
}

bool Collision::orientedBoundingBoxCollides(const Collider &boundingBox, const Collider &boundingBox2) {
    Vector3 axes[3] = {
            Vector3(1, 0, 0),
            Vector3(0, 1, 0),
            Vector3(0, 0, 1)
    };

    for (const auto& axis : axes) {
        float projection1Min = FLT_MAX, projection1Max = -FLT_MAX;
        float projection2Min = FLT_MAX, projection2Max = -FLT_MAX;

        for (const auto& vertex : boundingBox.vertices) {
            Vector3 vert = {vertex.x, vertex.y, vertex.z};
            float projection = vert.Dot(axis);
            if (projection < projection1Min) projection1Min = projection;
            if (projection > projection1Max) projection1Max = projection;
        }

        for (const auto& vertex : boundingBox2.vertices) {
            Vector3 vert = {vertex.x, vertex.y, vertex.z};
            float projection = vert.Dot(axis);
            if (projection < projection2Min) projection2Min = projection;
            if (projection > projection2Max) projection2Max = projection;
        }

        if (projection1Max < projection2Min || projection2Max < projection1Min)
            return false;
    }

    for (const auto& axis : axes) {
        float projection1Min = FLT_MAX, projection1Max = -FLT_MAX;
        float projection2Min = FLT_MAX, projection2Max = -FLT_MAX;

        for (const auto& vertex : boundingBox.vertices) {
            Vector3 vert = {vertex.x, vertex.y, vertex.z};
            float projection = vert.Dot(axis);
            if (projection < projection1Min) projection1Min = projection;
            if (projection > projection1Max) projection1Max = projection;
        }

        for (const auto& vertex : boundingBox2.vertices) {
            Vector3 vert = {vertex.x, vertex.y, vertex.z};
            float projection = vert.Dot(axis);
            if (projection < projection2Min) projection2Min = projection;
            if (projection > projection2Max) projection2Max = projection;
        }

        if (projection1Max < projection2Min || projection2Max < projection1Min)
            return false;
    }

    return true; //OBBs collide.
}

bool Collision::axisAlignedBoundingBoxCollides(const Collider &boundingBox, const Collider &boundingBox2) {
    Vector3 aabb1Min = {boundingBox.vertices.front().x, boundingBox.vertices.front().y, boundingBox.vertices.front().z};
    Vector3 aabb1Max = {boundingBox.vertices.back().x, boundingBox.vertices.back().y, boundingBox.vertices.back().z};
    Vector3 aabb2Min = {boundingBox2.vertices.front().x, boundingBox2.vertices.front().y, boundingBox2.vertices.front().z};
    Vector3 aabb2Max = {boundingBox2.vertices.back().x, boundingBox2.vertices.back().y, boundingBox2.vertices.back().z};

    if (aabb1Max.x < aabb2Min.x || aabb2Max.x < aabb1Min.x)
        return false;

    if (aabb1Max.y < aabb2Min.y || aabb2Max.y < aabb1Min.y)
        return false;

    if (aabb1Max.z < aabb2Min.z || aabb2Max.z < aabb1Min.z)
        return false;

    return true; //AABBs Collide.
}

//Broken
bool Collision::triangleVSAxisAlignedBoundingBoxCollides(const Triangle &triangle, const Collider &boundingBox) {
    return false;
}

bool Collision::collisionMapVSAxisAlignedBoundingBoxCollides(const Collider &boundingBox, const Collider& collisionMap) {
    for (auto& triangle : collisionMap.getTriangles())
        if (triangleVSAxisAlignedBoundingBoxCollides(triangle, boundingBox))
            return true;
    return false;
}

void Collider::draw() {
    if (type != ColliderType::CollisionMap) {
        glBegin(GL_LINES);
        for (int i = 0; i < 4; ++i) {
            glVertex3f(vertices[i].x, vertices[i].y, vertices[i].z);
            glVertex3f(vertices[(i + 1) % 4].x, vertices[(i + 1) % 4].y, vertices[(i + 1) % 4].z);

            glVertex3f(vertices[i + 4].x, vertices[i + 4].y, vertices[i + 4].z);
            glVertex3f(vertices[((i + 1) % 4) + 4].x, vertices[((i + 1) % 4) + 4].y, vertices[((i + 1) % 4) + 4].z);

            glVertex3f(vertices[i].x, vertices[i].y, vertices[i].z);
            glVertex3f(vertices[i + 4].x, vertices[i + 4].y, vertices[i + 4].z);
        }
        glEnd();
        return;
    }

    if (type == ColliderType::CollisionMap) {
        drawWireframe();

        for (const auto& triangle : getTriangles()) {
            Vector3 edge1 = triangle.V1 - triangle.V0;
            Vector3 edge2 = triangle.V2 - triangle.V0;
            Vector3 normal = Vector3::Normalize(Vector3::Cross(edge1, edge2));

            Vector3 centroid = (triangle.V0 + triangle.V1 + triangle.V2) / 3.0f;
            Vector3 normalPoint = centroid + (normal * 0.25);

            glColor3f(0.0f, 1.0f, 0.0f);
            glBegin(GL_LINES);
                glVertex3f(centroid.x, centroid.y, centroid.z);
                glVertex3f(normalPoint.x, normalPoint.y, normalPoint.z);
            glEnd();
            glColor4f(1.f, 1.f, 1.f, 1.f);
        }
    }

}

std::vector<Triangle> Collider::getTriangles() const {
    std::vector<Triangle> triangles;
    for (size_t i = 0; i < indices.size(); i += 3) {
        if (i + 2 < indices.size()) {
            Triangle triangle;
            triangle.V0 = vertices[indices[i]];
            triangle.V1 = vertices[indices[i + 1]];
            triangle.V2 = vertices[indices[i + 2]];
            triangles.push_back(triangle);
        }
    }
    return triangles;
}
