#include <engine/engine.h>
#include <glad/glad.h>
#include <sstream>
#include <engine/world.h>
#include <thread>
#include <types/entity/camera.h>
#include <types/entity/skybox.h>

#include <GL/glu.h>

using namespace J3ML;

void Engine::quit() const {
    window->destroyWindow();
    exit(0);
}

void outputErrorCode() {
    switch (engine->getError().error) {
        case ENGINE_ERROR_CODE::NO_ERROR:
            std::cerr << "NO ERROR" << std::endl;
            break;
        case ENGINE_ERROR_CODE::ANIM_OUT_OF_BOUNDS:
            std::cerr << "ANIM OUT OF BOUNDS" << std::endl;
            break;
        case ENGINE_ERROR_CODE::TEXTURE_NOT_FOUND:
            std::cerr << "TEXTURE NOT FOUND" << std::endl;
            break;
        case ENGINE_ERROR_CODE::SHADER_COMPILATION_ERROR:
            std::cerr << "SHADER COMPILATION ERROR" << std::endl;
            break;
        case ENGINE_ERROR_CODE::SHADER_LINKING_ERROR:
            std::cerr << "SHADER LINKING ERROR" << std::endl;
            break;
        case ENGINE_ERROR_CODE::INVALID_SHADER_TYPE:
            std::cerr << "INVALID SHADER TYPE" << std::endl;
            break;
        case ENGINE_ERROR_CODE::SHADER_NOT_FOUND:
            std::cerr << "SHADER NOT FOUND" << std::endl;
            break;
        case ENGINE_ERROR_CODE::MODEL_NOT_FOUND:
            std::cerr << "MODEL NOT FOUND" << std::endl;
            break;
        case ENGINE_ERROR_CODE::MULTI_TEXTURE_SIZE_EXCEEDS:
            std::cerr << "MULTI-TEXTURE SIZE EXCEEDS" << std::endl;
            break;
        case ENGINE_ERROR_CODE::TEXTURE_ERASED_WHILE_IN_USE:
            std::cerr << "TEXTURE ERASED WHILE IN USE" << std::endl;
    }
}

float Engine::getGLVersion()
{
    const std::string full_gl_ver = (const char*) (glGetString(GL_VERSION));
    const std::string gl_major = full_gl_ver.substr(0, 3);
    return std::stof(gl_major);
}

void Engine::initGL()
{
    gladLoadGL();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glOrtho(-1.0f, 1.0f, -1.0f, 1.0f, -0.5f, 0.5f);
    auto window_size = window->getSize();
    auto width_over_height = (float) window_size[0] / (float) window_size[1];
    auto aspect = width_over_height;
    gluPerspective(this->fov, aspect, this->nearPlane, this->farPlane);
    this->glError = glGetError();

    if (glError != GL_NO_ERROR) {
        std::cerr << "OpenGL: " << gluErrorString(glError);
        exit (1);
    }

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glViewport(0,0,window->getSize().x,window->getSize().y);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_TEXTURE_2D);
    glDepthMask(GL_TRUE);
    glClearDepth(1.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_VERTEX_ARRAY);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    //glEnable(GL_ALPHA_TEST);
    //glAlphaFunc(GL_GREATER, 0.05f); //Anything with an opacity of less than 5% gets thrown out
    //glEnable(GL_COLOR_MATERIAL);
    //glDepthRange(-0.5f,0.5f);
}

    void Engine::init()
    {
        engine->window->Open();
        engine->initGL();
        engine->loadConfig();
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    }

    void Engine::preRender()
    {
        // TODO: Move into Update(dt)
        // TODO: Create FixedUpdate(step)


        //std::cout << engine->frameCount << std::endl;
        engine->window->pollEvents();
        engine->frameCount++;
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        for (auto& e : engine->world->GetChildren())
            e->pre_render();
    }



    void Engine::render()
    {
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glCullFace(GL_BACK);
        for (auto& e : engine->world->GetChildren()) {
            if (e->draw)
                e->render();
        }
        engine->glError = glGetError();
        if (engine->glError != GL_NO_ERROR) {
            std::cerr << "OpenGL: " << gluErrorString(engine->glError) << std::endl;
            exit(0);
        }
    }

    void Engine::postRender()
    {
        for (auto& e : engine->world->GetChildren())
            e->post_render();

        //Reset all the transformations for the next frame.
        glPushMatrix();
        glPopMatrix();
        ReWindow::RWindow::glSwapBuffers();
        if (engine->frameCount == 1000)
            engine->takeScreenshot();
    }


    void Engine::renderPass()
    {
        preRender();
        render();
        postRender();
    }

    void Engine::renderLoop()
    {
        while (true) {
            auto start = std::chrono::high_resolution_clock::now();
            renderPass();
            auto stop = std::chrono::high_resolution_clock::now();
            float dT = (std::chrono::duration_cast<std::chrono::microseconds>(stop - start).count());

            //If we have more than 1000 fps.
            if (dT < 1000 && engine->window->getFlag(RWindowFlags::IN_FOCUS)) {
                int remaining = 1000 - dT;
                std::this_thread::sleep_for(std::chrono::microseconds(remaining));
                dT += remaining;
            }
            //Decrease framerate when focus is lost.
            if (!engine->window->getFlag(RWindowFlags::IN_FOCUS)) {
                int remaining = 41666.67f - dT;
                std::this_thread::sleep_for(std::chrono::microseconds(remaining));
                dT += remaining;
            }
            engine->frameDelta = (dT / 1000000);
        }
    }

float Engine::framerate() const
{
    return 1.f / this->frameDelta;
}

void Engine::takeScreenshot() {
    int width = (int) engine->window->getSize().x;
    int height = (int) engine->window->getSize().y;
    std::vector<unsigned char> frameBuffer(3 * width * height);

    glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, frameBuffer.data());

    std::string fileName = std::to_string(engine->frameCount) + ".bmp";
    std::ofstream file(fileName, std::ios::out | std::ios::binary);

    int fileSize = 54 + 3 * width * height;
    unsigned char bmpfileheader[14] = {'B', 'M', 0, 0, 0, 0, 0, 0, 0, 0, 54, 0, 0, 0};
    unsigned char bmpinfoheader[40] = {40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 24, 0};
    bmpfileheader[2] = (unsigned char) (fileSize);
    bmpfileheader[3] = (unsigned char) (fileSize >> 8);
    bmpfileheader[4] = (unsigned char) (fileSize >> 16);
    bmpfileheader[5] = (unsigned char) (fileSize >> 24);
    bmpinfoheader[4] = (unsigned char) (width);
    bmpinfoheader[5] = (unsigned char) (width >> 8);
    bmpinfoheader[6] = (unsigned char) (width >> 16);
    bmpinfoheader[7] = (unsigned char) (width >> 24);
    bmpinfoheader[8] = (unsigned char) (height);
    bmpinfoheader[9] = (unsigned char) (height >> 8);
    bmpinfoheader[10] = (unsigned char) (height >> 16);
    bmpinfoheader[11] = (unsigned char) (height >> 24);

    file.write((char*)bmpfileheader, 14);
    file.write((char*)bmpinfoheader, 40);

    int padding = (4 - (width * 3) % 4) % 4;
    for (int y = height - 1; y >= 0; --y) {
        for (int x = 0; x < width; ++x) {
            unsigned char pixel[3];
            pixel[0] = frameBuffer[(y * width + x) * 3 + 2];
            pixel[1] = frameBuffer[(y * width + x) * 3 + 1];
            pixel[2] = frameBuffer[(y * width + x) * 3];
            file.write((char*)pixel, 3);
        }
        for (int p = 0; p < padding; ++p) {
            file.write((char*)"\0", 1);
        }
    }

    file.close();
}
void Engine::loadConfig() {
    std::ifstream file("cfg/engine.cfg");
    if (!file.is_open()) {
        std::cerr << "Couldn't load engine config." << std::endl;
        this->quit();
        return;
    }
    std::string line;
    while (std::getline(file, line)) {
        std::istringstream stream(line);
        std::string prefix;
        stream >> prefix;
        //if (prefix == "Resolution:") {
            //int x, y;
            //stream >> x >> y;
            //window->setSize(x,y);
            //gluPerspective(this->fov, x / y, this->nearPlane, this->farPlane);
            //glViewport(0,0,x,y);
        //}
        if (prefix == "Fullscreen:")
            stream >> fullscreen;
        if (prefix == "Debug:")
            stream >> debug;
        if (prefix == "CameraFOV:")
            stream >> fov;
    }
    //If we have a window already.
    //if (window != NULL) {
        //window->destroyWindow();
        //Some window managers won't handle destroying and remaking the window with zero delay very well.
        //std::this_thread::sleep_for(std::chrono::milliseconds(500));
        //initVideo();
        //initGL();
    //}
    std::cout << "Fullscreen: " << fullscreen << std::endl;
    std::cout << "Resolution: " << window->getSize()[0] << "x" << window->getSize()[1] << std::endl;
    std::cout << "Camera FOV: " << fov << std::endl;
    std::cout << "Debug: " << debug << std::endl;
}

void Engine::setError(ENGINE_ERROR_CODE code, bool critical) {
    error.error = code;
    error.critical = critical;
    outputErrorCode();
    if (error.critical)
        engine->quit(code);

}

EngineError Engine::getError() {
    return error;
}

void Engine::quit(ENGINE_ERROR_CODE code) {
    window->destroyWindow();
    exit((int) code);
}
