#include <engine/engine.h>
#include <types/vertex.h>
#include <engine/collision.h>
#include <cmath>
#include <tuple>

/// Reads a single Vertex Coordinate Attribute from an OBJ format text file.
// v = vertex
Vector3 ParseVertexCoordinates(std::istringstream& stream) {

	float x, y, z;
	stream >> x >> y >> z;
	return {x,y,z};
}

/// Reads a single Texture Coordinate Attribute from an OBJ format text file.
// vt = vertex texture coordinates
Vector2 ParseVertexTextureCoordinates(std::istringstream& stream) {
	float u, v;
	stream >> u >> v;
	return {u,v};
}


/// Reads a single Face Index Attribute from an OBJ format text file.
/// This tells us what order to place our list of vertices and texcoords to create
/// our model.
// f = face
FaceIndices ParseFaceIndices(std::istringstream& stream) {
	//unsigned int vertexIndex[3], texCoordIndex[3];
	FaceIndices face;
        char slash;
        for (int i = 0; i < 3; ++i) {
		unsigned int vI;
		unsigned int tCI;

		// Get vertexIndex value seperated by slash and get texcoordindex value
		stream >> vI >> slash >> tCI;

		face.vertexIndex[i] = vI;
		face.texCoordIndex[i] = tCI;

		// Shift vertexindex by 1
		face.vertexIndex[i]--;
		// shift texcoordindex by 1
		face.texCoordIndex[i]--;
	}

	return face;
}


// map face of model to positions and texture coords
void VertexArray::mapfaces(const FaceIndices& facedata, const std::vector<Vector3>& positions, const std::vector<Vector2>& uvs) {
    	for (int i = 0; i < 3; ++i) {
		Vertex vertex;
		TextureCoordinate textureCoordinate;
		vertex = positions[facedata.vertexIndex[i]];
		textureCoordinate = uvs[facedata.texCoordIndex[i]];

		vertices.push_back(vertex);
		texCoords.push_back(textureCoordinate);
		indices.push_back(static_cast<unsigned int>(indices.size()));
	}
}


void VertexArray::load (const std::string& filename) {
    std::ifstream file(filename);
    if (!file.is_open())
        engine->setError(ENGINE_ERROR_CODE::MODEL_NOT_FOUND, true);

    std::vector<Vector3> positions;
    std::vector<Vector2> uvs;
    std::string line;

    // First: We read every (unique) vertex entry and textureCoord entry from the file.
    // Then: Use Face data to map the bulk data into correct locations.

    // TODO: Check the assumption that we can **always** read all "v", "vt" etc. tokens  first,
    // allowing us to perform a second-pass to reconstruct faces?

    while (std::getline(file, line)) {
        std::istringstream stream(line);
        std::string prefix;
        stream >> prefix;


        if (prefix == "v") {
            auto result =  ParseVertexCoordinates(stream);
            positions.emplace_back(result);
        } else if (prefix == "vt") {
            auto result = ParseVertexTextureCoordinates(stream);
            uvs.emplace_back(result);
        } else if (prefix == "f") {
            auto result = ParseFaceIndices(stream);
            mapfaces(result, positions, uvs);
        }
    }
    file.close();

    //Copy verts to vram.
    if (engine->useVBO) {
        if (vbo == 0) {
            //This breaks stuff if the model is *very* big.
            GLfloat verts[vertices.size() * 3];
            size_t index = 0;
            for (const auto &vertex: vertices) {
                verts[index++] = vertex.x;
                verts[index++] = vertex.y;
                verts[index++] = vertex.z;
            }
            glGenBuffers(1, &vbo);
            glBindBuffer(GL_ARRAY_BUFFER, vbo);
            glBufferData(GL_ARRAY_BUFFER, sizeof(verts), &verts, GL_STATIC_DRAW); //GL_STATIC_DRAW means the VBO won't be changed atall.
            glBindBuffer(GL_ARRAY_BUFFER, 0);

            //Copy Texture Coordinates to vram.
            GLfloat textureCoordinates[texCoords.size() * 2];
            index = 0;
            for (const auto &tCoord: texCoords) {
                textureCoordinates[index++] = tCoord.x;
                textureCoordinates[index++] = tCoord.y;
            }
            glGenBuffers(1, &tbo);
            glBindBuffer(GL_ARRAY_BUFFER, tbo);
            glBufferData(GL_ARRAY_BUFFER, sizeof(textureCoordinates), &textureCoordinates, GL_STATIC_DRAW);
            glBindBuffer(GL_ARRAY_BUFFER, 0);

            //copy indices to vram.
            glGenBuffers(1, &ebo);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo);
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, (indices.size() * sizeof(GLuint)), indices.data(), GL_STATIC_DRAW);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        }
    }

    //TODO: This will probably need to be changed when pill collider & skeletal animation is added.
    //Create OBB once when object is loaded because recalculating it every frame is a LOT slower than rotating it every frame.
    boundingVolume = Collision::calculateOrientedBoundingBox(*this, {0,0,0}).vertices;

}

void VertexArray::draw() {
    if (!indices.empty()) {
        glEnable(GL_TEXTURE_2D);
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        if (!engine->useVBO) {
            glVertexPointer(3, GL_FLOAT, sizeof(Vertex), &vertices[0].x);
            glTexCoordPointer(2, GL_FLOAT, sizeof(TextureCoordinate), &texCoords[0].x);
            glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &indices[0]);
        }

        if (engine->useVBO) {
            //Vertices
            glBindBuffer(GL_ARRAY_BUFFER, vbo);
            glVertexPointer(3, GL_FLOAT, 0, nullptr);

            //Texture Coordinates
            glBindBuffer(GL_ARRAY_BUFFER, tbo);
            glTexCoordPointer(2, GL_FLOAT, 0, nullptr);

            //Indices
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo);
            glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, nullptr);

            //Reset bound buffer.
            glBindBuffer(GL_ARRAY_BUFFER, 0);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        }

        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisable(GL_TEXTURE_2D);
    }
}

void VertexArray::drawWireframe() {
    glBegin(GL_LINES);
    for (size_t i = 0; i < indices.size(); i += 2) {
        if (i + 1 < indices.size()) {
            int index1 = static_cast<int>(indices[i]);
            int index2 = static_cast<int>(indices[i + 1]);

            if (index1 >= 0 && index1 < vertices.size() && index2 >= 0 && index2 < vertices.size()) {
                glVertex3f(vertices[index1].x, vertices[index1].y, vertices[index1].z);
                glVertex3f(vertices[index2].x, vertices[index2].y, vertices[index2].z);
            }
        }
    }
    glEnd();
}

std::tuple<float, float> ComputeVertexRoll(Vertex vertex, float roll) {
	float newX = vertex.x * std::cos(roll) - vertex.y * std::sin(roll);
	float newY = vertex.x * std::sin(roll) + vertex.y * std::cos(roll);
	return {newX, newY};
}

std::tuple<float, float> ComputeVertexYaw(Vertex vertex, float yaw) {
	float newX = vertex.x * std::cos(yaw) + vertex.z * std::sin(yaw);
        float newZ = -vertex.x * std::sin(yaw) + vertex.z * std::cos(yaw);
	return {newX, newZ};
}

std::tuple<float, float> ComputeVertexPitch(Vertex vertex, float pitch) {
        float newY = vertex.y * std::cos(pitch) - vertex.z * std::sin(pitch);
        float newZ = vertex.y * std::sin(pitch) + vertex.z * std::cos(pitch);
	return {newY, newZ};
}



void VertexArray::rotate(const Vector3& angle) {
    float angleX = Math::Radians(angle.x);
    float angleY = Math::Radians(angle.y);
    float angleZ = Math::Radians(angle.z);

    // Roll
    for (Vertex &vertex: vertices) {
        /*float newX = vertex.x * std::cos(angleZ) - vertex.y * std::sin(angleZ);
        float newY = vertex.x * std::sin(angleZ) + vertex.y * std::cos(angleZ);*/
	auto result = ComputeVertexRoll(vertex, angleZ);
        vertex.x = std::get<0>(result);
        vertex.y = std::get<1>(result);
    }

    // Yaw
    for (Vertex &vertex: vertices) {
        /*float newX = vertex.x * std::cos(angleY) + vertex.z * std::sin(angleY);
        float newZ = -vertex.x * std::sin(angleY) + vertex.z * std::cos(angleY);*/
	auto result = ComputeVertexYaw(vertex, angleY);
        vertex.x = std::get<0>(result);
        vertex.z = std::get<1>(result);
    }

    // Pitch
    for (Vertex &vertex: vertices) {
        /*float newY = vertex.y * std::cos(angleX) - vertex.z * std::sin(angleX);
        float newZ = vertex.y * std::sin(angleX) + vertex.z * std::cos(angleX);*/
	auto result = ComputeVertexPitch(vertex, angleX);
        vertex.y = std::get<0>(result);
        vertex.z = std::get<1>(result);
    }
}

VertexArray VertexArray::getBoundingVolume(Vector3 angle) {
    VertexArray vArray;
    vArray.vertices = boundingVolume;
    vArray.rotate(angle);
    return vArray;

}

void VertexArray::erase(const VertexArray& vArray) {
    if (engine->useVBO) {
        glDeleteBuffers(1, &vArray.vbo);
        glDeleteBuffers(1, &vArray.tbo);
        glDeleteBuffers(1, &vArray.ebo);
    }

    for (int i = 0; i < engine->world->geometryList.size(); i++) {
        if (engine->world->geometryList[i].name == vArray.name)
            engine->world->geometryList.erase(engine->world->geometryList.begin() + i);
    }
}

VertexArray::VertexArray(const std::string &filename) {
    load(filename);
}
